import uuid
import json

CURRENT_FORMAT_VERSION = "2"


class NotImplementedException(BaseException):
    pass


def get_attribute(attribute_name, data):
    return data["attributes"][attribute_name]


def get_object_from_relationship(included, key):
    for item in included:
        if item["type"] == key["type"] and item["id"] == key["id"]:
            return item


class IncludedObject:
    def __init__(self, included_data):
        self._data = {}
        for item in included_data:
            self.include(item)

    def include(self, item):
        type_ = item["type"]
        id_ = item["id"]

        if type_ not in self._data:
            self._data[type_] = {}
        if id_ not in self._data[type_]:
            self._data[type_][id_] = item

    def get_object_from_relationship(self, key):
        return self._data[key["type"]][key["id"]]

    def get_data_for_relationship(self, relationship_name, data):
        return self.get_object_from_relationship(data["relationships"][relationship_name]["data"])

    @property
    def as_list(self):
        included = []
        for type_ in self._data:
            for id_ in self._data[type_]:
                included.append(self._data[type_][id_])
        return included


class JSONAPIMixin:
    @classmethod
    def create_from_json_api_string(cls, json_string):
        raise NotImplementedException("The method 'create_from_json_api_string' is not implemented.")

    @classmethod
    def create_from_json_api_data(cls, json_string):
        raise NotImplementedException("The method 'create_from_json_api_data' is not implemented.")

    @property
    def as_json_api_string(self):
        obj_as_dict = self.as_json_api_dict()

        obj_as_dict["included"] = sorted(obj_as_dict["included"], key=lambda d: f"{d['type']}{d['id']}")

        return json.dumps(obj_as_dict)
        # return json.dumps(obj_as_dict, indent=4)

    def as_json_api_dict(self):
        def get_relationship(name):
            target = self.__getattribute__(name)
            try:
                return target.as_json_api_dict()
            except AttributeError:
                return None

        obj = JSONAPIObject(self.json_api_type, self.json_api_id, {}, {}, [])
        for attribute in self.json_api_attributes:
            obj.add_attribute(attribute, self.__getattribute__(attribute))
        for relationship_name in self.json_api_relationships:
            relationship = get_relationship(relationship_name)
            if relationship:
                obj.add_relationship(relationship_name, relationship)
        return obj.as_dict

    @property
    def json_api_type(self):
        type_ = f"{self.__class__.__module__}.{self.__class__.__name__}"
        return type_

    @property
    def json_api_id(self):
        return self.json_api_type

    @property
    def json_api_attributes(self):
        return []

    @property
    def json_api_relationships(self):
        return []


class JSONAPIObject:
    @classmethod
    def create_from_dict(cls, json_data):
        data = json_data["data"]
        included = json_data["included"]

        return cls(data["type"], data["id"], data["attributes"], data["relationships"], included)

    def __init__(self, type_, id_=None, attributes=None, relationships=None, included=None):
        self.type_ = type_
        self.id_ = id_ if id_ else str(uuid.uuid4())
        self.attributes = attributes if attributes else {}
        self.relationships = relationships if relationships else {}

        included = included if included else []
        self.included = IncludedObject(included)

    def set_id(self, id_):
        self.id_ = id_

    def set_type(self, type_):
        self.type_ = type_

    def add_relationship(self, name, entry):
        if entry == []:
            self.relationships[name] = {"data": []}
        else:
            self.relationships[name] = {"data": {"type": entry["data"]["type"], "id": entry["data"]["id"]}}
            if isinstance(entry["included"], dict):
                self.include(entry["included"])
            else:
                for item in entry["included"]:
                    self.include(item)
            self.include(entry["data"])

    def add_relationship_no_include(self, name, entry):
        self.relationships[name] = {"data": {"type": entry["data"]["type"], "id": entry["data"]["id"]}}

    def append_relationship(self, name, entry):
        if name not in self.relationships:
            self.relationships[name] = {"data": []}

        self.relationships[name]["data"].append({"type": entry["data"]["type"], "id": entry["data"]["id"]})
        for item in entry["included"]:
            self.include(item)
        self.include(entry["data"])

    def has_relationship(self, name, id_):
        if name in self.relationships:
            if type(self.relationships[name]["data"] is list):
                for item in self.relationships[name]["data"]:
                    if item["id"] == id_:
                        return True
                return False
            else:
                return self.relationships[name]["data"]["id"] == id_

    def add_attribute(self, name, entry):
        self.attributes[name] = entry

    def append_attribute(self, name, entry):
        if name not in self.attributes:
            self.attributes[name] = []

        self.attributes[name].append(entry)

    def include(self, data):
        if data:
            self.included.include(data)

    @property
    def as_dict(self):
        return {
            "data": {
                "type": self.type_,
                "id": self.id_,
                "version:id": "2",
                "attributes": self.attributes,
                "relationships": self.relationships
            },
            "included": self.included.as_list
        }

    @property
    def as_json(self):
        return json.dumps(self.as_dict, indent=4)



import threading
import json

import paho.mqtt.client as mqtt

from tala.utils import sse_client


class MQTTClientException(BaseException):
    pass


class MQTTClient(sse_client.AbstractSSEClient):
    sessions = {}

    def __init__(self, client_id_base, logger, endpoint, port=None):
        super().__init__(client_id_base, logger, endpoint, port)

        def on_connect(client, userdata, connect_flags, reason_code, properties):
            self.logger.info('CONNACK received', reason_code=reason_code, properties=properties)
            self._connected.set()

        self._connected = threading.Event()

        self._client = mqtt.Client(
            mqtt.CallbackAPIVersion.VERSION2,
            transport="websockets",
            reconnect_on_failure=True,
            clean_session=True,
            client_id=self._client_id
        )
        self._client.on_connect = on_connect
        self._client.tls_set()

    def start(self):
        self.logger.info("connecting to", endpoint=self._endpoint, port=self._port)
        self._client.connect(self._endpoint, self._port)
        self._client.loop_start()

    def _get_topic(self, streamer_session):
        return f'tm/id/{streamer_session["session_id"]}'

    def _stream_to_frontend(self, streamer_session, message):
        def remove_final_space(message):
            if "data" in message:
                message["data"] = message["data"].strip()
            return message

        streamer_session["message_counter"] += 1
        self.logger.debug(
            "MQTT Client streaming to frontend",
            message=message,
            session_id=streamer_session["session_id"],
            client_id=self.client_id
        )
        self._connected.wait()
        self._client.publish(self._get_topic(streamer_session), json.dumps(message))

    def close_session(self, session_id):
        streamer_session = self.sessions[session_id]
        streamer_exception = streamer_session.get("streaming_exception")
        if streamer_exception:
            raise MQTTClientException(
                f'{streamer_exception} was raised during streaming in {streamer_session["session_id"]}. Streamed: {self._streamed}.'
            )
        self.logger.info("close session", client_id=self.client_id, session_id=streamer_session["session_id"])
        self.logger.info(
            "Streamed in session",
            num_messages=streamer_session["message_counter"],
            streamed=streamer_session["streamed"]
        )
        self._reset_logger()
        del self.sessions[session_id]


class ChunkJoiner(sse_client.ChunkJoiner):
    pass


class StreamIterator(sse_client.StreamIterator):
    pass



import unittest

from tala import model

from tala.model.plan import Plan, UnableToDetermineOntologyException
from tala.model.plan_item import IfThenElse, Findout, Raise, AssumeShared
from tala.model.semantic_object import OntologySpecificSemanticObject
from tala.testing.lib_test_case import LibTestCase
from tala.utils.json_api import IncludedObject


class PlanTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.maxDiff = None
        self.findout_price = Findout(self.domain_name, self.price_question)
        self.findout_dest_city = Findout(self.domain_name, self.dest_city_question)
        self.raise_price = Raise(self.domain_name, self.price_question)
        self.raise_dest_city = Raise(self.domain_name, self.dest_city_question)
        self.consequent = [self.raise_price]
        self.alternative = [self.raise_dest_city]
        self.if_then_else_item = IfThenElse("mockup_condition", self.consequent, self.alternative)
        self.assume_shared = AssumeShared(self.proposition_dept_city_london)

    def test_plan_iteration_steps_into_nested_blocks(self):
        self._given_plan_with_nested_item()
        expected_list = [self.findout_price, self.findout_dest_city, self.consequent[0], self.alternative[0]]
        self.assertEqual(expected_list, list(self.plan))

    def _given_plan_with_nested_item(self):
        self.plan = Plan([self.if_then_else_item, self.findout_dest_city, self.findout_price])

    def test_plan_iteration_after_modification(self):
        plan = Plan()
        plan.push(self.findout_price)
        expected_list = [self.findout_price]
        self.assertEqual(expected_list, list(plan))

    def test_removal_of_findout_from_nested_block(self):
        self._given_plan_with_nested_item()
        self._when_consequent_is_removed()
        self._then_all_elements_but_consequent_are_left()

    def _when_consequent_is_removed(self):
        self.plan.remove(self.consequent[0])

    def _then_all_elements_but_consequent_are_left(self):
        expected_list = [self.findout_price, self.findout_dest_city, self.alternative[0]]
        self.assertEqual(expected_list, list(self.plan))

    def test_removal_of_plan_item_from_deeply_nested_block(self):
        self._given_plan_with_deeply_nested_item()
        self._when_consequent_is_removed()
        self._then_plan_contains([self.findout_price, self.findout_dest_city, self.alternative[0]])

    def _given_plan_with_deeply_nested_item(self):
        self.nested_if_then_else = IfThenElse("outer_condition", [self.if_then_else_item], [])
        self.plan = Plan([self.raise_dest_city, self.findout_dest_city, self.findout_price])

    def _then_plan_contains(self, expected_list):
        self.assertEqual(expected_list, list(self.plan))

    def test_plan_iteration_steps_into_very_deeply_nested_blocks(self):
        self.maxDiff = None
        self._given_plan_with_very_deeply_nested_item()
        expected_list = [
            self.findout_price,
            self.findout_dest_city,
            self.raise_price,
            self.raise_dest_city,
            self.findout_dest_city,
        ]
        self.assertEqual(expected_list, list(self.plan))

    def _given_plan_with_very_deeply_nested_item(self):
        self.nested_if_then_else = IfThenElse("outer_condition", [self.if_then_else_item], [self.findout_dest_city])
        self.doubly_nested_if_then_else = IfThenElse(
            "outer_condition", [self.nested_if_then_else], [self.findout_price]
        )
        self.plan = Plan([self.nested_if_then_else, self.findout_dest_city, self.findout_price])

    def test_plan_as_json_api_base_case(self):
        self.given_plan([self.findout_price])
        self.when_plan_as_json_api()
        self.then_plan_as_json_api_is({
            'data': {
                'attributes': {
                    'accommodate_without_feedback': [],
                    'alternatives_predicate': [],
                    'max_answers': [],
                    'ontology_name': 'mock_ontology',
                    'reraise_on_resume': [],
                    'restart_on_completion': [],
                    'unrestricted_accommodation': [],
                    'version:id': '2'
                },
                'id': 'mock_ontology:PERFORM_GOAL:mock_action',
                'relationships': {
                    'goal': {
                        'data': {
                            'id': 'PERFORM_GOAL.SYS:mock_action',
                            'type': 'tala.model.goal.Perform'
                        }
                    },
                    'plan_content': {
                        'data': [{
                            'id': 'tala.model.plan_item.Findout:mockup_ontology:WHQ:X.price(X):False',
                            'type': 'tala.model.plan_item.Findout'
                        }]
                    }
                },
                'type': 'tala.model.plan',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'name': 'mock_action',
                    'ontology_name': 'mock_ontology'
                },
                'id': 'mock_ontology:mock_action',
                'type': 'tala.model.action'
            }, {
                'attributes': {
                    'target': 'SYS',
                    'type_': 'PERFORM_GOAL'
                },
                'id': 'PERFORM_GOAL.SYS:mock_action',
                'relationships': {
                    'content': {
                        'data': {
                            'id': 'mock_ontology:mock_action',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.goal.Perform',
                'version:id': '2'
            }, {
                'attributes': {},
                'id': 'real',
                'relationships': {},
                'type': 'tala.model.sort.RealSort',
                'version:id': '2'
            }, {
                'attributes': {
                    '_multiple_instances': False,
                    'feature_of_name': None,
                    'name': 'price',
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'mockup_ontology:price',
                'relationships': {
                    'sort': {
                        'data': {
                            'id': 'real',
                            'type': 'tala.model.sort.RealSort'
                        }
                    }
                },
                'type': 'tala.model.predicate.Predicate',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'X.price(X)',
                'relationships': {
                    'predicate': {
                        'data': {
                            'id': 'mockup_ontology:price',
                            'type': 'tala.model.predicate.Predicate'
                        }
                    }
                },
                'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology',
                    'type_': 'WHQ'
                },
                'id': 'mockup_ontology:WHQ:X.price(X)',
                'relationships': {
                    'content': {
                        'data': {
                            'id': 'X.price(X)',
                            'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition'
                        }
                    }
                },
                'type': 'tala.model.question.WhQuestion',
                'version:id': '2'
            }, {
                'attributes': {
                    'allow_answer_from_pcom': False,
                    'domain_name': 'mockup_domain',
                    'type_': 'findout'
                },
                'id': 'tala.model.plan_item.Findout:mockup_ontology:WHQ:X.price(X):False',
                'relationships': {
                    'question': {
                        'data': {
                            'id': 'mockup_ontology:WHQ:X.price(X)',
                            'type': 'tala.model.question.WhQuestion'
                        }
                    }
                },
                'type': 'tala.model.plan_item.Findout',
                'version:id': '2'
            }]
        })

    def given_plan(self, plan_item_list):
        self._plan = Plan(plan_item_list)

    def when_plan_as_json_api(self):
        goal = model.goal.Perform(model.action.Action("mock_action", "mock_ontology"))

        self._plan_as_json_api = self._plan.as_json_api_dict(goal, [], [], [], [], [], [], [], [], [], [])

    def then_plan_as_json_api_is(self, result):
        self.assertEqual(result, self._plan_as_json_api)

    def test_plan_as_json_api_more_items(self):
        self.given_plan([self.findout_price, self.raise_dest_city])
        self.when_plan_as_json_api()
        self.then_plan_as_json_api_is({
            'data': {
                'attributes': {
                    'accommodate_without_feedback': [],
                    'alternatives_predicate': [],
                    'max_answers': [],
                    'ontology_name': 'mock_ontology',
                    'reraise_on_resume': [],
                    'restart_on_completion': [],
                    'unrestricted_accommodation': [],
                    'version:id': '2'
                },
                'id': 'mock_ontology:PERFORM_GOAL:mock_action',
                'relationships': {
                    'goal': {
                        'data': {
                            'id': 'PERFORM_GOAL.SYS:mock_action',
                            'type': 'tala.model.goal.Perform'
                        }
                    },
                    'plan_content': {
                        'data': [{
                            'id': 'tala.model.plan_item.Findout:mockup_ontology:WHQ:X.price(X):False',
                            'type': 'tala.model.plan_item.Findout'
                        }, {
                            'id': 'tala.model.plan_item.Raise:mockup_ontology:WHQ:X.dest_city(X):False',
                            'type': 'tala.model.plan_item.Raise'
                        }]
                    }
                },
                'type': 'tala.model.plan',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'name': 'mock_action',
                    'ontology_name': 'mock_ontology'
                },
                'id': 'mock_ontology:mock_action',
                'type': 'tala.model.action'
            }, {
                'attributes': {
                    'target': 'SYS',
                    'type_': 'PERFORM_GOAL'
                },
                'id': 'PERFORM_GOAL.SYS:mock_action',
                'relationships': {
                    'content': {
                        'data': {
                            'id': 'mock_ontology:mock_action',
                            'type': 'tala.model.action'
                        }
                    }
                },
                'type': 'tala.model.goal.Perform',
                'version:id': '2'
            }, {
                'attributes': {},
                'id': 'real',
                'relationships': {},
                'type': 'tala.model.sort.RealSort',
                'version:id': '2'
            }, {
                'attributes': {
                    '_multiple_instances': False,
                    'feature_of_name': None,
                    'name': 'price',
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'mockup_ontology:price',
                'relationships': {
                    'sort': {
                        'data': {
                            'id': 'real',
                            'type': 'tala.model.sort.RealSort'
                        }
                    }
                },
                'type': 'tala.model.predicate.Predicate',
                'version:id': '2'
            }, {
                'attributes': {
                    '_multiple_instances': False,
                    'feature_of_name': None,
                    'name': 'dest_city',
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'mockup_ontology:dest_city',
                'relationships': {
                    'sort': {
                        'data': {
                            'id': 'mockup_ontology:city',
                            'type': 'tala.model.sort.CustomSort'
                        }
                    }
                },
                'type': 'tala.model.predicate.Predicate',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'X.price(X)',
                'relationships': {
                    'predicate': {
                        'data': {
                            'id': 'mockup_ontology:price',
                            'type': 'tala.model.predicate.Predicate'
                        }
                    }
                },
                'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'X.dest_city(X)',
                'relationships': {
                    'predicate': {
                        'data': {
                            'id': 'mockup_ontology:dest_city',
                            'type': 'tala.model.predicate.Predicate'
                        }
                    }
                },
                'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology',
                    'type_': 'WHQ'
                },
                'id': 'mockup_ontology:WHQ:X.price(X)',
                'relationships': {
                    'content': {
                        'data': {
                            'id': 'X.price(X)',
                            'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition'
                        }
                    }
                },
                'type': 'tala.model.question.WhQuestion',
                'version:id': '2'
            }, {
                'attributes': {
                    'ontology_name': 'mockup_ontology',
                    'type_': 'WHQ'
                },
                'id': 'mockup_ontology:WHQ:X.dest_city(X)',
                'relationships': {
                    'content': {
                        'data': {
                            'id': 'X.dest_city(X)',
                            'type': 'tala.model.lambda_abstraction.LambdaAbstractedPredicateProposition'
                        }
                    }
                },
                'type': 'tala.model.question.WhQuestion',
                'version:id': '2'
            }, {
                'attributes': {
                    'allow_answer_from_pcom': False,
                    'domain_name': 'mockup_domain',
                    'type_': 'findout'
                },
                'id': 'tala.model.plan_item.Findout:mockup_ontology:WHQ:X.price(X):False',
                'relationships': {
                    'question': {
                        'data': {
                            'id': 'mockup_ontology:WHQ:X.price(X)',
                            'type': 'tala.model.question.WhQuestion'
                        }
                    }
                },
                'type': 'tala.model.plan_item.Findout',
                'version:id': '2'
            }, {
                'attributes': {
                    'dynamic': True,
                    'name': 'city',
                    'ontology_name': 'mockup_ontology'
                },
                'id': 'mockup_ontology:city',
                'relationships': {},
                'type': 'tala.model.sort.CustomSort',
                'version:id': '2'
            }, {
                'attributes': {
                    'allow_answer_from_pcom': False,
                    'domain_name': 'mockup_domain',
                    'type_': 'raise'
                },
                'id': 'tala.model.plan_item.Raise:mockup_ontology:WHQ:X.dest_city(X):False',
                'relationships': {
                    'question': {
                        'data': {
                            'id': 'mockup_ontology:WHQ:X.dest_city(X)',
                            'type': 'tala.model.question.WhQuestion'
                        }
                    }
                },
                'type': 'tala.model.plan_item.Raise',
                'version:id': '2'
            }]
        })

    def test_plan_created_from_plan_as_json_api(self):
        self.given_plan([self.assume_shared, self.findout_price, self.findout_dest_city])
        self.given_plan_as_json_api()
        self.given_plan_created_from_json_api()
        self.then_created_plan_contains_same_elements_as_old_plan()

    def given_plan_as_json_api(self):
        goal = model.goal.Perform(model.action.Action("mock_action", "mock_ontology"))

        self._plan_as_json_api = self._plan.as_json_api_dict(goal, [], [], [], [], 0, [], [], [], [], [])

    def given_plan_created_from_json_api(self):
        self._plan_from_json_api = Plan.create_from_json_api_data(
            self._plan_as_json_api["data"], IncludedObject(self._plan_as_json_api["included"])
        )[1]

    def then_created_plan_contains_same_elements_as_old_plan(self):
        for original, created in zip(self._plan.iter_stack(), self._plan_from_json_api.iter_stack()):
            self.assertEqual(original, created)


class SemanticObjectPlanTests(unittest.TestCase):
    def setUp(self):
        self._semantic_objects = set()

    def test_ontology_name_with_ontology_specific_semantic_object(self):
        self._given_semantic_object_of_ontology("an ontology")
        self._given_plan()
        self._when_asking_for_ontology_name()
        self._then_ontology_name_is("an ontology")

    def _given_semantic_object_of_ontology(self, ontology):
        semantic_object = OntologySpecificSemanticObject(ontology)
        self._semantic_objects.add(semantic_object)

    def _given_plan(self):
        self._plan = Plan(self._semantic_objects)

    def _when_asking_for_ontology_name(self):
        self._result = self._plan.ontology_name

    def _then_ontology_name_is(self, expected_name):
        actual_name = self._result
        self.assertEqual(expected_name, actual_name)

    def test_ontology_name_with_multiple_ontology_specific_semantic_objects(self):
        self._given_semantic_object_of_ontology("an ontology")
        self._given_semantic_object_of_ontology("another ontology")
        self._given_plan()
        self._when_asking_for_ontology_name_then_exception_is_raised()

    def _when_asking_for_ontology_name_then_exception_is_raised(self):
        with self.assertRaises(UnableToDetermineOntologyException):
            self._result = self._plan.ontology_name

    def test_ontology_name_with_multiple_ontology_specific_semantic_objects_of_same_ontology(self):
        self._given_semantic_object_of_ontology("an ontology")
        self._given_semantic_object_of_ontology("an ontology")
        self._given_plan()
        self._when_asking_for_ontology_name()
        self._then_ontology_name_is("an ontology")



import collections

from tala.utils.as_json import AsJSONMixin
from tala.utils.json_api import JSONAPIMixin

SEMANTIC_OBJECT_TYPE = "ask_feature"


class AskFeature(collections.namedtuple("AskFeature", ["name", "kpq"]), AsJSONMixin, JSONAPIMixin):
    @classmethod
    def create_from_json_api_dict(cls, ask_feature_data, included):
        name = ask_feature_data["attributes"]["name"]
        kpq = ask_feature_data["attributes"]["kpq"]
        return cls(name, kpq)

    def __new__(cls, predicate_name, kpq=False):
        return super(AskFeature, cls).__new__(cls, predicate_name, kpq)

    def as_dict(self):
        return {"semantic_object_type": SEMANTIC_OBJECT_TYPE, "name": self.name, "kpq": self.kpq}

    @property
    def json_api_id(self):
        return f"{self.name}:{self.kpq}"

    @property
    def json_api_attributes(self):
        return ["name", "kpq"]

    @property
    def json_api_relationships(self):
        return []



import unittest

from unittest.mock import Mock

from tala.model.proposition import Proposition
from tala.model.validation_result import ValidationResult, ValidationSuccess, ValidationFailure


class TestBase(unittest.TestCase):
    def setUp(self):
        self._validation_object = None
        self._result = None

    def when_accessing_is_successful(self):
        self._result = self._validation_object.is_successful

    def then_result_is(self, expected_result):
        self.assertEqual(self._result, expected_result)

    def when_calling_repr(self):
        self._result = repr(self._validation_object)


class ValidationResultTests(TestBase):
    def test_creation(self):
        self.given_created_object()
        self.when_accessing_is_successful_then_exception_is_raised_matching(
            NotImplementedError, "This property needs to be implemented in a subclass"
        )

    def given_created_object(self):
        self._validation_object = ValidationResult()

    def when_accessing_is_successful_then_exception_is_raised_matching(self, expected_exception, expected_message):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._result = self._validation_object.is_successful


class ValidationSuccessTests(TestBase):
    def test_is_successful_property_of_successful_result(self):
        self.given_created_object()
        self.when_accessing_is_successful()
        self.then_result_is(True)

    def given_created_object(self):
        self._validation_object = ValidationSuccess()

    def test_repr_of_successful_result(self):
        self.given_created_object()
        self.when_calling_repr()
        self.then_result_is("ValidationSuccess()")

    def test_eq_on_identical_objects(self):
        self.when_equaling(ValidationSuccess(), ValidationSuccess())
        self.then_result_is(True)

    def test_eq_on_instance_of_other_class(self):
        self.when_equaling(ValidationSuccess(), "non_validation_result")
        self.then_result_is(False)

    def when_equaling(self, first, second):
        self._result = first == second

    def test_ne_on_identical_objects(self):
        self.when_testing_not_equals(ValidationSuccess(), ValidationSuccess())
        self.then_result_is(False)

    def when_testing_not_equals(self, first, second):
        self._result = first != second

    def test_ne_on_instance_of_other_class(self):
        self.when_testing_not_equals(ValidationSuccess(), "non_validation_result")
        self.then_result_is(True)

    def test_ne_on_failure(self):
        self.when_testing_not_equals(ValidationSuccess(), ValidationFailure("a_reason", Mock(spec=Proposition)))
        self.then_result_is(True)

    def test_uniqueness_of_identical_objects(self):
        self.when_checking_length_of({ValidationSuccess(), ValidationSuccess()})
        self.then_result_is(1)

    def when_checking_length_of(self, validation_results):
        self._result = len(validation_results)

    def test_uniqueness_of_different_objects(self):
        self.when_checking_length_of({ValidationSuccess(), ValidationFailure("a_reason", Mock(spec=Proposition))})
        self.then_result_is(2)


class ValidationFailureTests(TestBase):
    def setUp(self):
        super(ValidationFailureTests, self).setUp()
        self._invalid_parameters = None
        self._first = None
        self._second = None

    def test_is_successful_property_of_failed_result(self):
        self.given_created_failed_result_with(invalidity_reason="a_reason", invalid_parameters="mock_parameter_set")
        self.when_accessing_is_successful()
        self.then_result_is(False)

    def given_created_failed_result_with(self, invalidity_reason, invalid_parameters=None):
        self._validation_object = self._create_failed_result(invalidity_reason, invalid_parameters)

    def _create_failed_result(self, invalidity_reason, invalid_parameters=None):
        invalid_parameters = invalid_parameters or "mock_parameter_set"
        return ValidationFailure(invalidity_reason=invalidity_reason, invalid_parameters=invalid_parameters)

    def test_repr_of_failed_result(self):
        self.given_created_failed_result_with("a_reason", "mock_parameter_set")
        self.when_calling_repr()
        self.then_result_is("ValidationFailure('a_reason', 'mock_parameter_set')")

    def test_invalidity_reason_property(self):
        self.given_created_failed_result_with("a_reason")
        self.when_accessing_invalidity_reason()
        self.then_result_is("a_reason")

    def when_accessing_invalidity_reason(self):
        self._result = self._validation_object.invalidity_reason

    def test_invalid_parameters_property(self):
        self.given_created_failed_result_with("a_reason", "mock_parameter_set")
        self.when_accessing_invalid_parameters()
        self.then_result_is("mock_parameter_set")

    def when_accessing_invalid_parameters(self):
        self._result = self._validation_object.invalid_parameters

    def test_eq_on_identical_objects(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second(self._create_failed_result("a_reason"))
        self.when_equaling_first_and_second()
        self.then_result_is(True)

    def given_first(self, first):
        self._first = first

    def given_second(self, second):
        self._second = second

    def given_created_two_identical_validation_failures(self):
        self._first = self._create_failed_result("a_reason")
        self._second = self._create_failed_result("a_reason")

    def when_equaling_first_and_second(self):
        self._result = self._first == self._second

    def test_eq_on_instance_of_other_class(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second("mock_non_validation_result")
        self.when_equaling_first_and_second()
        self.then_result_is(False)

    def test_ne_on_identical_objects(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second(self._create_failed_result("a_reason"))
        self.when_testing_not_equals_on_first_and_second()
        self.then_result_is(False)

    def when_testing_not_equals_on_first_and_second(self):
        self._result = self._first != self._second

    def test_ne_on_instance_of_other_class(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second("mock_non_validation_result")
        self.when_testing_not_equals_on_first_and_second()
        self.then_result_is(True)

    def test_ne_on_success(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second(ValidationSuccess())
        self.when_testing_not_equals_on_first_and_second()
        self.then_result_is(True)

    def test_uniqueness_of_identical_objects(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second(self._create_failed_result("a_reason"))
        self.when_checking_length_of_first_and_second()
        self.then_result_is(1)

    def when_checking_length_of_first_and_second(self):
        self._result = len({self._first, self._second})

    def test_uniqueness_of_different_objects(self):
        self.given_first(self._create_failed_result("a_reason"))
        self.given_second(self._create_failed_result("another_reason"))
        self.when_checking_length_of_first_and_second()
        self.then_result_is(2)



# encoding: utf-8

import pytest
from unittest.mock import patch, Mock

import tala.utils.as_json
from tala.utils.as_json import AsJSONMixin, convert_to_json
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin


class TestConvertToJson(object):
    def test_none_is_returned_as_is(self):
        self.given_object(None)
        self.when_convert_to_json()
        self.then_result_is(None)

    def given_object(self, object_):
        self._object = object_

    def when_convert_to_json(self, verbose=True):
        self._result = convert_to_json(self._object, verbose=verbose)

    def then_result_is(self, expected):
        assert expected == self._result

    @pytest.mark.parametrize("value", [True, False])
    def test_boolean_is_returned_as_is(self, value):
        self.given_object(value)
        self.when_convert_to_json()
        self.then_result_is(value)

    def test_as_json_mixin_is_processed_with_its_as_dict_method(self):
        self.given_a_mocked_as_json_mixin(as_dict_return_value={"key": "value"})
        self.when_convert_to_json()
        self.then_result_is({"key": "value"})

    def given_a_mocked_as_json_mixin(self, *args, **kwargs):
        self._object = self._create_mocked_as_json_mixin(*args, **kwargs)

    def _create_mocked_as_json_mixin(self, as_dict_return_value=None):
        mock = Mock(spec=AsJSONMixin)
        mock.as_dict.return_value = as_dict_return_value or {}
        return mock

    def test_list_is_processed_recursively_and_returned_as_list(self):
        self.given_a_mocked_as_json_mixin_as_a_list(as_dict_return_value={"key": "value"})
        self.when_convert_to_json()
        self.then_result_is([{"key": "value"}])

    def given_a_mocked_as_json_mixin_as_a_list(self, *args, **kwargs):
        self._object = [self._create_mocked_as_json_mixin(*args, **kwargs)]

    def test_set_is_processed_recursively_and_returned_as_dict(self):
        self.given_a_mocked_as_json_mixin_as_a_set(as_dict_return_value={"key": "value"})
        self.when_convert_to_json()
        self.then_result_is({"set": [{"key": "value"}]})

    def given_a_mocked_as_json_mixin_as_a_set(self, *args, **kwargs):
        self._object = {self._create_mocked_as_json_mixin(*args, **kwargs)}

    def test_dict_is_processed_recursively_and_returned_as_dict(self):
        self.given_object({"key": self._create_mocked_as_json_mixin(as_dict_return_value={"inner_key": "inner_value"})})
        self.when_convert_to_json()
        self.then_result_is({"key": {"inner_key": "inner_value"}})

    def test_that_non_verbose_call_with_as_json_is_processed_with_its_as_dict_method(self):
        self.given_a_mocked_as_json_mixin(as_dict_return_value={"key": "value"})
        self.when_convert_to_json(verbose=False)
        self.then_result_is({"key": "value"})

    def test_non_verbose_call_with_as_semantic_expression_mixin(self):
        self.given_a_mocked_as_semantic_expression_mixin(as_semantic_expression_return_value="expression")
        self.when_convert_to_json(verbose=False)
        self.then_result_is({"semantic_expression": "expression"})

    def given_a_mocked_as_semantic_expression_mixin(self, as_semantic_expression_return_value):
        mock = Mock(spec=AsSemanticExpressionMixin)
        mock.as_semantic_expression.return_value = as_semantic_expression_return_value
        self._object = mock

    def test_non_verbose_call_with_object_mixed_in_with_json_and_semantic_expression(self):
        self.given_a_mocked_object_mixed_in_with_both_json_and_semantic_expression(
            as_semantic_expression_return_value="expression", as_dict_return_value={"key": "value"}
        )
        self.when_convert_to_json(verbose=False)
        self.then_result_is({
            "semantic_expression": "expression",
        })

    def given_a_mocked_object_mixed_in_with_both_json_and_semantic_expression(
        self, as_semantic_expression_return_value, as_dict_return_value
    ):
        class JSONAndSemanticExpressionMixin(AsJSONMixin, AsSemanticExpressionMixin):
            pass

        mock = Mock(spec=JSONAndSemanticExpressionMixin)
        mock.as_semantic_expression.return_value = as_semantic_expression_return_value
        mock.as_dict.return_value = as_dict_return_value
        self._object = mock

    def test_verbose_call_with_object_mixed_in_with_json_and_semantic_expression(self):
        self.given_a_mocked_object_mixed_in_with_both_json_and_semantic_expression(
            as_semantic_expression_return_value="expression", as_dict_return_value={"key": "value"}
        )
        self.when_convert_to_json(verbose=True)
        self.then_result_is({
            "key": "value",
            "semantic_expression": "expression",
        })

    def test_unicode_of_object_is_returned_as_fallback(self):
        self.given_an_object_that_should_be_treated_with_a_fallback_strategy(unicode_return_value="åäö unicode_string")
        self.when_convert_to_json()
        self.then_result_is("åäö unicode_string")

    def given_an_object_that_should_be_treated_with_a_fallback_strategy(self, unicode_return_value):
        class MockCustomClass(object):
            def __str__(self):
                return unicode_return_value

        self._object = MockCustomClass()


class TestAsJSONMixin(object):
    def setup_method(self):
        self._mock_convert_to_json = None
        self._as_json_mixin = None
        self._result = None

    @patch("{}.convert_to_json".format(tala.utils.as_json.__name__), autospec=True)
    def test_as_json_returns_convert_to_json_result(self, mock_convert_to_json):
        self.given_mock_convert_to_json(mock_convert_to_json)
        self.given_as_json_mixin()
        self.when_call_as_json()
        self.then_result_is_from_mocked_convert_to_json()

    def given_mock_convert_to_json(self, mock_convert_to_json):
        self._mock_convert_to_json = mock_convert_to_json

    def given_as_json_mixin(self):
        self._as_json_mixin = AsJSONMixin()

    def when_call_as_json(self):
        self._result = self._as_json_mixin.as_json()

    def then_result_is_from_mocked_convert_to_json(self):
        assert self._result == self._mock_convert_to_json.return_value

    @patch("{}.convert_to_json".format(tala.utils.as_json.__name__), autospec=True)
    def test_as_json_calls_convert_to_json_with_verbose(self, mock_convert_to_json):
        self.given_mock_convert_to_json(mock_convert_to_json)
        self.given_as_json_mixin()
        self.when_call_as_json()
        self.then_convert_to_json_is_called_with_as_json_mixin_object(verbose=True)

    def then_convert_to_json_is_called_with_as_json_mixin_object(self, verbose):
        self._mock_convert_to_json.assert_called_once_with(self._as_json_mixin, verbose=verbose)

    @patch("{}.convert_to_json".format(tala.utils.as_json.__name__), autospec=True)
    def test_as_compact_json_returns_dict_with_values_converted_to_json(self, mock_convert_to_json):
        self.given_mock_convert_to_json(mock_convert_to_json)
        self.given_as_json_mixin()
        self.when_call_as_compact_json()
        self.then_result_is_from_mocked_convert_to_json()

    def when_call_as_compact_json(self):
        self._result = self._as_json_mixin.as_compact_json()

    @patch("{}.convert_to_json".format(tala.utils.as_json.__name__), autospec=True)
    def test_as_compact_json_calls_convert_to_json_without_verbose(self, mock_convert_to_json):
        self.given_mock_convert_to_json(mock_convert_to_json)
        self.given_as_json_mixin()
        self.when_call_as_compact_json()
        self.then_convert_to_json_is_called_with_as_json_mixin_object(verbose=False)



import re
import random
from string import Formatter

from tala.utils.func import log_models
from tala.utils.compression import ensure_decompressed_json

PROTOCOL_VERSION = "1.0"

SUCCESS = "success"
FAIL = "fail"
ERROR = "error"
VALID_STATUSES = [SUCCESS, FAIL, ERROR]

RERAISE_SV = "Så "
RERAISE_EN = "So "

ICM_SLOT_PATTERN = r'icm:per\*pos\:(.*)'
PROPOSITIONAL_SLOT_PATTERN = r'(\$[a-zA-Z0-9_]+)'
STRING_ANSWER_PATTERN = r'[a-zA-Z0-9_]+\(\"([^"]*)\"\)'
INDIVIDUAL_SLOT = "&individual"
PREDICATE_WILDCARD = "*predicate"
PROPOSITION_SET_WILDCARD = "*proposition_set"
VALIDITY_WILDCARD = "*validity"


class SlotDefinitionException(Exception):
    def __init__(self, message):
        super().__init__()
        self._message = message

    @property
    def message(self):
        return self._message


class NoMoveSequenceFoundException(Exception):
    pass


def get_predicate(proposition_expression):
    m = re.match(r"([a-zA-Z0-9_]+)\(.*", proposition_expression)
    if m:
        return m.groups()[0]


def list_of_strings_to_string(list_of_strings):
    def quote_string(string):
        return "'" + string + "'"

    def quote_strings(list_of_strings):
        return [quote_string(string) for string in list_of_strings]

    return "[" + (", ".join(quote_strings(list_of_strings))) + "]"


def is_utterance_with_ng_slots(utterance):
    fields = list(Formatter().parse(utterance))
    if len(fields) > 1:
        return True
    return False


def generate(moves, context, session, logger):
    moves = [{"semantic_expression": move} for move in moves]

    request = {"moves": moves, "context": context, "session": session}
    result = nlg(request, logger)
    logger.info("generate() returns", result=result)
    return result


def generate_utterance(moves, context, session, logger):
    result = generate(moves, context, session, logger)
    if result["status"] == SUCCESS:
        return result["utterance"]
    return ""


def generate_moves_subsequences(moves):
    for i in range(0, len(moves) - 1):
        yield list_of_strings_to_string(moves[i:])


def nlg(body, logger):
    def get_nlg_model():
        try:
            if body.get("nlg"):
                logger.debug("Collecting NLG data from the request")
                nlg_model = body["nlg"]
            if "session" in body and body["session"].get("nlg"):
                logger.debug("Collecting NLG data from the session object")
                nlg_model = body["session"].get("nlg")

            return ensure_decompressed_json(nlg_model)
        except UnboundLocalError as e:
            logger.warning("No NLG model could be found in the request body or the session object", error=e)

    def lowercase_question_with_reraise_sw_en(utterance):
        for target_token in [RERAISE_EN, RERAISE_SV]:
            if target_token in utterance:
                index = utterance.find(target_token) + len(target_token)
                utterance = utterance[:index] + utterance[index].lower() + utterance[index + 1:]
                return utterance
        return utterance

    def decide_persona(utterances):
        for utterance in utterances:
            persona = utterance.get("persona", "tutor")
            if persona is not None:
                return persona
        return None

    def get_voice(nlg_data, utterance_persona):
        persona = nlg_data["personas"].get(utterance_persona, {})
        return persona.get("voice")

    def make_and_log_response(response):
        try:
            if "_" in response["utterance"]:
                logger.warning("response contains an underscore character: _", utterance=response["utterance"])
        except Exception:
            pass
        logger.info("responding", response=response)
        return response

    def get_facts():
        try:
            return body["context"]["facts"]
        except (KeyError, TypeError):
            return {}

    def get_facts_being_grounded():
        return body.get("context", {}).get("facts_being_grounded", {})

    def get_entities_under_discussion():
        return body.get("context", {}).get("entities_under_discussion", {})

    log_models(body, logger, ["nlg"])
    logger.info("incoming NLG request", body=body)
    moves = [move["semantic_expression"] for move in body["moves"]]

    nlg_data = get_nlg_model()
    if not nlg_data:
        return make_and_log_response(
            response={
                "status": FAIL,
                "message": "No NLG model could be found in the request body or the session object"
            }
        )

    g = Generator(nlg_data, get_facts(), get_facts_being_grounded(), get_entities_under_discussion(), logger)

    try:
        utterance = g.generate_sequence(moves)
        persona = utterance.get("persona", "tutor")
        return make_and_log_response(
            response={
                "status": SUCCESS,
                "utterance": utterance["utterance"],
                "persona": persona,
                "voice": get_voice(nlg_data, persona)
            }
        )
    except NoMoveSequenceFoundException:
        pass
    try:
        utterances = list(g.generate(move) for move in moves)
        logger.debug("utterances", utterances=utterances)
        clean_utterances = [utterance for utterance in utterances if utterance["utterance"]]
        final_utterance = " ".join(utterance["utterance"] for utterance in clean_utterances)
        potentially_lowercased_utterance = lowercase_question_with_reraise_sw_en(final_utterance)
        logger.debug(
            "potentially_lowercased_utterance", potentially_lowercased_utterance=potentially_lowercased_utterance
        )
        if potentially_lowercased_utterance == "":
            return make_and_log_response(
                response={
                    "status": FAIL,
                    "message": f"moves {moves} was generated as the empty string."
                }
            )
    except SlotDefinitionException as exception:
        logger.warning(exception.message)
        return make_and_log_response(response={"status": FAIL, "message": exception.message})
    persona = decide_persona(utterances)
    return make_and_log_response(
        response={
            "status": SUCCESS,
            "utterance": potentially_lowercased_utterance,
            "persona": persona,
            "voice": get_voice(nlg_data, persona)
        }
    )


class Generator:
    def __init__(self, nlg_data, facts, facts_being_grounded, entities_under_discussion, logger):
        self._facts = facts
        self._nlg_data = nlg_data
        self._facts_being_grounded = facts_being_grounded
        self._entities_under_discussion = entities_under_discussion
        self._logger = logger

    def generate_sequence(self, moves):
        for moves_as_string in generate_moves_subsequences(moves):
            sequence_content = self._nlg_data.get(moves_as_string)
            if sequence_content:
                utterance = self._select_candidate_utterance_from_string(sequence_content["utterance"])
                if is_utterance_with_ng_slots(utterance):
                    sequence_content["utterance"] = self._populate_ng_slots_in(utterance)
                else:
                    sequence_content["utterance"] = utterance
                return sequence_content
        raise NoMoveSequenceFoundException(f"no sequence matching '{moves}' found")

    def generate(self, move):
        def _is_move_in_patterns_with_exact_match(nlg_data_doc):
            if nlg_data_doc:
                return True
            return False

        nlg_data_doc = self._nlg_data.get(move)
        if _is_move_in_patterns_with_exact_match(nlg_data_doc):
            if is_utterance_with_ng_slots(nlg_data_doc["utterance"]):
                utterance = self._handle_utterance_with_ng_slots(nlg_data_doc["utterance"])
            else:
                utterance = self._handle_utterance_possibly_with_og_slots(nlg_data_doc["utterance"])
            if "_" in utterance:
                self._logger.warning("base case: move in mappings", utterance=utterance)
            return {"utterance": utterance, "persona": nlg_data_doc.get("persona")}

        self._logger.debug("move is not exact match, try slots.")
        slot_pattern = self._get_generalized_slot_pattern(move)
        self._logger.debug("slot_pattern", slot_pattern)
        nlg_data_doc = self._nlg_data.get(slot_pattern)

        if nlg_data_doc:
            self._logger.debug("calling _handle_utterance_with_ng_slots", nlg_data_doc["utterance"])
            utterance = self._handle_utterance_with_ng_slots(nlg_data_doc["utterance"])
            if "_" in utterance:
                self._logger.warning("base case: move in mappings", utterance=utterance)
            return {"utterance": utterance, "persona": nlg_data_doc.get("persona")}

        for population_function in [
            self._populate_predicate_and_validity_patterns,
            self._populate_proposition_set_patterns,
            self._populate_validity_patterns,
            self._populate_individual_slot_patterns,
            self._populate_propositional_slot_patterns,
            self._populate_icm_references,
            self._get_string_from_string_answer_move,
        ]:
            result = population_function(move)
            if result and result["utterance"] is not None:
                if "_" in result:
                    self._logger.warning(f"{population_function.__name__} applied to utterance", utterance=result)
                return result
        self._logger.warning(f"The move '{move}' was not found in the database. Generating the empty string.")

        return {"utterance": "", "persona": None}

    def _get_generalized_slot_pattern(self, move):
        self._logger.debug("_get_generalized_slot_pattern for move", move)
        m = re.match(r'(answer\([a-zA-Z0-9_]+\()[a-zA-Z0-9_]+(\)\))', move)
        if m:
            replacement = m.group(1) + "*" + m.group(2)
            self._logger.debug("replacing", move, "with", replacement)
            return replacement

    def _handle_utterance_possibly_with_og_slots(self, utterance_candidates):
        utterance = self._select_candidate_utterance_from_string(utterance_candidates)
        return self._populate_slots_in(utterance)

    def _handle_utterance_with_ng_slots(self, utterance_candidates):
        utterance = self._select_candidate_utterance_from_string(utterance_candidates)
        return self._populate_ng_slots_in(utterance)

    def _populate_ng_slots_in(self, utterance):
        def all_facts_dict():
            return self._facts | self._facts_being_grounded | self._entities_under_discussion

        def create_filler_dict():
            d = {}
            all_facts = all_facts_dict()
            for predicate_name in all_facts:
                d[predicate_name] = grammar_entry(predicate_name, all_facts)
            return d

        def grammar_entry(entry, all_facts):
            if all_facts[entry].get("grammar_entry", None):
                return all_facts[entry].get("grammar_entry", None)
            if all_facts[entry]["sort"] in ["integer", "real", "string"]:
                return all_facts[entry].get("value")
            individual = all_facts[entry].get("value", None)
            result = self.generate(f"answer({individual})")
            return result["utterance"]

        filler_dict = create_filler_dict()
        return utterance.format_map(filler_dict)

    def _select_candidate_utterance_from_string(self, candidates):
        candidate_utterances = candidates.split("|")
        return self._select_candidate_from_utterances(candidate_utterances)

    def _select_candidate_from_utterances(self, utterances):
        selected_utterance = random.choice(utterances)
        stripped_utterance = selected_utterance.strip()
        return stripped_utterance

    def _populate_slots_in(self, utterance):
        def get_predicate_from_slot(slot_definition):
            m = re.match(r"@([a-zA-Z0-9_]+)", slot_definition)
            if m:
                return m.groups()[0]

        def get_slot_definition(utterance):
            m = re.search(r'(@[a-zA-Z0-9_]+)', utterance)
            if m:
                match = m.groups()[0]
                return match
            return None

        slot_definition = get_slot_definition(utterance)
        if slot_definition is None:
            return utterance
        predicate = get_predicate_from_slot(slot_definition)
        grammar_entry = self._get_grammar_entry_for(predicate)
        return utterance.replace(slot_definition, grammar_entry)

    def _get_matcher_pattern_for(self, move_pattern, slot_definitions):
        replacement = "(.*)"
        escaped_move_pattern = re.escape(move_pattern)
        matcher_pattern = escaped_move_pattern
        for slot_definition in slot_definitions:
            escaped_slot_definition = re.escape(slot_definition)
            matcher_pattern = matcher_pattern.replace(escaped_slot_definition, replacement)
        return matcher_pattern

    def _get_grammar_entry_for(self, predicate):
        for collection in [self._facts, self._facts_being_grounded, self._entities_under_discussion]:
            if predicate in collection:
                if collection[predicate]["sort"] == "string":
                    return collection[predicate]["value"]
                if collection[predicate].get("grammar_entry", None) is not None:
                    return collection[predicate]["grammar_entry"]
                return collection[predicate]["value"]
        self._logger.warning(f"Expected predicate with entry in context, but got '{predicate}'.")
        return ""

    def _replace_slot(self, utterance_pattern, slot_reference, grammar_entry):
        utterance = utterance_pattern.replace(slot_reference, grammar_entry)
        return utterance

    def _pattern_matches_move(self, pattern, move, slot_references):
        re_pattern = self._get_matcher_pattern_for(pattern, slot_references)
        match_object = re.match(re_pattern, move)
        return match_object

    def _populate_predicate_and_validity_patterns(self, move):
        for entry in self._nlg_data["validity_wildcard_entries"]["docs"]:
            if VALIDITY_WILDCARD in entry["match"] and PREDICATE_WILDCARD in entry["match"]:
                matches_predicate = self._pattern_matches_move(
                    entry["match"], move, [PREDICATE_WILDCARD, VALIDITY_WILDCARD]
                )
                if matches_predicate:
                    return {"utterance": entry["utterance"], "persona": entry.get("persona")}

        for entry in self._nlg_data["predicate_wildcard_entries"]["docs"]:
            if VALIDITY_WILDCARD in entry["match"] and PREDICATE_WILDCARD in entry["match"]:
                matches_predicate = self._pattern_matches_move(
                    entry["match"], move, [PREDICATE_WILDCARD, VALIDITY_WILDCARD]
                )
                if matches_predicate:
                    return {"utterance": entry["utterance"], "persona": entry.get("persona")}

    def _populate_validity_patterns(self, move):
        for entry in self._nlg_data["validity_wildcard_entries"]["docs"]:
            if VALIDITY_WILDCARD in entry["match"]:
                if self._pattern_matches_move(entry["match"], move, [VALIDITY_WILDCARD]):
                    return {"utterance": entry["utterance"], "persona": entry.get("persona")}

    def _populate_proposition_set_patterns(self, move):
        for entry in self._nlg_data["proposition_set_wildcard_entries"]["docs"]:
            if PROPOSITION_SET_WILDCARD in entry["match"]:
                if self._pattern_matches_move(entry["match"], move, [PROPOSITION_SET_WILDCARD]):
                    return {"utterance": entry["utterance"], "persona": entry.get("persona")}

    def _populate_individual_slot_patterns(self, move):
        def populate_pattern_with_individual_slots(utterance_pattern, move_pattern, propositional_slot):
            re_pattern = self._get_matcher_pattern_for(move_pattern, [propositional_slot])
            match_object = re.match(re_pattern, move)
            if match_object:
                proposition_expression = match_object.group(1)
                predicate = get_predicate(proposition_expression)
                grammar_entry = self._get_grammar_entry_for(predicate)
                result = self._replace_slot(utterance_pattern, INDIVIDUAL_SLOT, grammar_entry)
                return result

        def get_propositional_slot_reference_for_individual(pattern):
            m = re.search(r'([a-zA-Z0-9_]+\(' + INDIVIDUAL_SLOT + r'\))', pattern)
            if m:
                match = m.groups()[0]
                return match
            return None

        for entry in self._nlg_data["individual_entries"]["docs"]:
            utterance_pattern = entry["utterance"]
            propositional_slot_for_individual = get_propositional_slot_reference_for_individual(entry["match"])
            if propositional_slot_for_individual is not None:
                if self._pattern_matches_move(entry["match"], move, ["&individual"]):
                    if INDIVIDUAL_SLOT in entry["match"] and INDIVIDUAL_SLOT not in utterance_pattern:
                        return {"utterance": utterance_pattern, "persona": entry.get("persona")}
                    result_utterance = populate_pattern_with_individual_slots(
                        utterance_pattern, entry["match"], propositional_slot_for_individual
                    )
                    return {"utterance": result_utterance, "persona": entry.get("persona")}

    def _populate_propositional_slot_patterns(self, move):
        def populate_pattern_with_propositional_slots(utterance_pattern, move_pattern, slot_definition):
            re_pattern = self._get_matcher_pattern_for(move_pattern, [slot_definition])
            match_object = re.match(re_pattern, move)
            if match_object:
                proposition_expression = match_object.group(1)
                predicate = get_predicate(proposition_expression)
                grammar_entry = self._get_grammar_entry_for(predicate)
                if grammar_entry == "":
                    self._logger.warning(f"Expected move with entry in context, but got '{move}'.")
                    return ""
                return self._replace_slot(utterance_pattern, slot_definition, grammar_entry)

        for entry in self._nlg_data["propositional_entries"]["docs"]:
            slot_reference = self._get_propositional_slot_reference_for(entry["match"])
            if slot_reference:
                utterance_pattern = entry["utterance"]
                if slot_reference in entry["match"] and slot_reference not in utterance_pattern:
                    raise SlotDefinitionException(
                        f"Expected '{entry['match']}' and '{utterance_pattern}' to contain same slot."
                    )
                if self._pattern_matches_move(entry["match"], move, [slot_reference]):
                    result_utterance = populate_pattern_with_propositional_slots(
                        utterance_pattern, entry["match"], slot_reference
                    )
                    return {"utterance": result_utterance, "persona": entry.get("persona")}

    def _get_propositional_slot_reference_for(self, pattern):
        m = re.search(PROPOSITIONAL_SLOT_PATTERN, pattern)
        if m:
            match = m.groups()[0]
            return match
        return None

    def _populate_icm_references(self, move):
        def get_icm_slot_reference_for(match):
            m = re.search(ICM_SLOT_PATTERN, match)
            if m:
                match = m.groups()[0]
                return match
            return None

        def populate_pattern_with_string(entry, icm_slot_reference, move):
            utterance_pattern = entry["utterance"]
            re_pattern = self._get_matcher_pattern_for(entry["match"], [icm_slot_reference])
            match_object = re.match(re_pattern, move)
            if match_object:
                embedded_string = match_object.group(1)
                result = self._replace_slot(utterance_pattern, icm_slot_reference, embedded_string)
                return result

        for entry in self._nlg_data["icm_slot_entries"]["docs"]:
            icm_slot_reference = get_icm_slot_reference_for(entry["match"])
            if icm_slot_reference is not None \
               and self._pattern_matches_move(entry["match"], move, [icm_slot_reference]):
                result_utterance = populate_pattern_with_string(entry, icm_slot_reference, move)
                return {"utterance": result_utterance, "persona": entry.get("persona")}

    def _get_string_from_string_answer_move(self, move):
        m = re.search(STRING_ANSWER_PATTERN, move)
        if m:
            match = m.groups()[0]
            return {"utterance": match, "persona": None}
        return None



from tala.utils.equality import EqualityMixin
from tala.utils.as_json import AsJSONMixin


class UnexpectedModalityException(Exception):
    pass


class Entity(EqualityMixin, AsJSONMixin):
    def __init__(self, name: str, sort: str, natural_language_form: str, ddd_name: str = None) -> None:
        self._name = name
        self._sort = sort
        self._natural_language_form = natural_language_form
        self._ddd_name = ddd_name

    @classmethod
    def create_from_json(cls, input):
        name = input["name"]
        sort = input["sort"]
        natural_language_form = input["natural_language_form"]
        ddd_name = input.get("ddd")
        return Entity(name, sort, natural_language_form, ddd_name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def sort(self) -> str:
        return self._sort

    @property
    def natural_language_form(self) -> str:
        return self._natural_language_form

    @property
    def ddd_name(self) -> str:
        return self._ddd_name

    def __str__(self):
        return "{}({}, {}, {}, {})".format(
            self.__class__.__name__, self._name, self._sort, self._natural_language_form, self._ddd_name
        )

    def __repr__(self):
        return str(self)

    def as_dict(self):
        if self.ddd_name:
            return {
                "name": self.name,
                "sort": self.sort,
                "natural_language_form": self.natural_language_form,
                "ddd": self.ddd_name,
            }
        else:
            return {
                "name": self.name,
                "sort": self.sort,
                "natural_language_form": self.natural_language_form,
            }



import os
import shutil
import sys
import tempfile
import unittest

from tala.config import BackendConfig, DddConfig


class DddMockingTestCase(unittest.TestCase):
    def setUp(self):
        self._working_dir = os.getcwd()
        temp_dir = tempfile.mkdtemp(prefix="DddMockingTestCase")
        os.chdir(temp_dir)
        self._temp_dir = os.getcwd()
        self._exception_occurred = False
        self._mock_ddd_config = DddConfig.default_config(use_rgl=True)
        self._mock_backend_config = BackendConfig.default_config(active_ddd="mockup_app", ddds=["mockup_app"])

    def tearDown(self):
        os.chdir(self._working_dir)
        if self._exception_occurred:
            if "--keep-temp-dir" in sys.argv:
                print(("NOTE: Temp dir %s kept for debugging. It needs to be deleted manually." % self._temp_dir))
            else:
                self._delete_temp_dir()
                print("NOTE: Temp dir deleted. To keep the temp dir for debugging, "
                      "run again with --keep-temp-dir")
        else:
            self._delete_temp_dir()

    def _delete_temp_dir(self):
        shutil.rmtree(self._temp_dir)

    def _given_ontology_py_file(self, path):
        content = """
from tala.model.ontology import DddOntology
class MockupOntology(DddOntology):
  sorts = {}
  predicates = {}
  individuals = {}
  actions = set()
"""
        self.create_mockup_file(path, content)

    def _given_domain_py_file(self, path):
        content = """
from tala.model.domain import DddDomain
class MockupDomain(DddDomain):
  plans = []
"""
        self.create_mockup_file(path, content)

    def create_mockup_domain_file(self, path="mockup_app/domain.xml"):
        content = """
<domain name="MockupDomain"/>
"""
        self.create_mockup_file(path, content)

    def _given_domain_xml_file(self, path, domain_name="MockupDomain"):
        content = '<domain name="%s"/>' % domain_name
        self.create_mockup_file(path, content)

    def _given_service_interface_xml_file(self, path):
        content = '<service_interface/>'
        self.create_mockup_file(path, content)

    def _given_ontology_xml_file(self, path):
        content = '<ontology name="MockupOntology"/>'
        self.create_mockup_file(path, content)

    def _given_mocked_ddd_config(self, **kwargs):
        if "use_rgl" not in kwargs:
            kwargs["use_rgl"] = True
        self._mock_ddd_config = DddConfig.default_config(**kwargs)

    def _given_mocked_backend_config(self, **kwargs):
        self._mock_backend_config = BackendConfig.default_config(**kwargs)

    def create_mockup_file(self, relative_path, content):
        container_directory = os.path.dirname(relative_path)
        self._ensure_dir_exists(container_directory)
        with open(relative_path, "w") as f:
            f.write(content)

    def _ensure_dir_exists(self, path):
        if not os.path.exists(path):
            os.makedirs(path)



from tala.log.extractor import LogExtractor, REQUEST_EVENT, RESPONSE_EVENT, SELECTED_INTERPRETATION_EVENT
from tala.testing.interactions.compiler import InteractionTestCompiler


class FailedException(Exception):
    pass


class UnexpectedInputException(Exception):
    pass


class InteractionTestExtractor:
    def __init__(self, log_file, print_mixin, verbose=False):
        self._log_file = log_file
        self._print_mixin = print_mixin
        self._verbose = verbose
        self._log_extractor = LogExtractor(self._log_file)

    @staticmethod
    def from_args(log, verbose=False, full=False, semantic=False):
        def create_mixin():
            if full:
                if semantic:
                    return FullSemanticMixin()
                return FullNaturalLanguageMixin()
            if semantic:
                return ReadableSemanticMixin()
            return ReadableNaturalLanguageMixin()

        args = (log, create_mixin(), verbose)
        return InteractionTestExtractor(*args)

    def run(self):
        if self._verbose:
            self._print_header()
        entries = list(self._log_extractor.get_backend_communication_entries())
        for index, entry in enumerate(entries):
            self._print_mixin.handle_entry(entries, entry, index)

    def _print_header(self):
        self._print_tdm_version()
        self._print_language()
        self._print_working_dir()
        self._print_config()
        print()

    def _print_tdm_version(self):
        logged_version = self._log_extractor.get_backend_tdm_version()
        self._print_header_entry("Originally run by TDM", logged_version)

    def _print_working_dir(self):
        logged_dir = self._log_extractor.get_backend_working_directory()
        self._print_header_entry("with working directory", logged_dir)

    def _print_language(self):
        logged_language = self._log_extractor.get_backend_language()
        self._print_header_entry("in language", logged_language)

    def _print_config(self):
        logged_backend_config = self._log_extractor.get_backend_config()
        self._print_header_entry("using backend config", logged_backend_config)

    def _print_header_entry(self, name, logged_value):
        if logged_value is None:
            logged_value = "Unknown"

        print(f"# {name} {logged_value}")


class PrintMixin:
    @classmethod
    def handle_entry(cls, entries, entry, index):
        raise NotImplementedError()


class FullSemanticMixin(PrintMixin):
    @classmethod
    def handle_entry(cls, entries, entry, index):
        if entry["event"] == REQUEST_EVENT:
            request = entry["frontend_request"]["request"]
            if "semantic_input" in request:
                print(InteractionTestCompiler.pretty_semantic_input(request["semantic_input"]))
            elif "natural_language_input" in request:
                FullNaturalLanguageMixin.handle_entry(entries, entry, index)
            elif "passivity" in request:
                print(InteractionTestCompiler.pretty_passivity())
            elif "event" in request:
                print(InteractionTestCompiler.pretty_event(request["event"]))
        if entry["event"] == RESPONSE_EVENT:
            if "output" in entry["frontend_response"]:
                moves = entry["frontend_response"]["output"]["moves"]
                print(InteractionTestCompiler.pretty_system_moves(moves))
        if entry["event"] == SELECTED_INTERPRETATION_EVENT:
            pass


class ReadableSemanticMixin(FullSemanticMixin):
    @classmethod
    def handle_entry(cls, entries, entry, index):
        def is_next_entry_more_informed():
            next_index = index + 1
            if len(entries) <= next_index:
                return False
            next_entry = entries[next_index]
            if next_entry["event"] == SELECTED_INTERPRETATION_EVENT:
                return next_entry["selected_interpretation"] is not None
            return False

        def average(elements):
            if not elements:
                return 0.0
            return sum(elements) / len(elements)

        def interpretation_confidence(interpretation):
            moves = interpretation["moves"]
            return average([move["perception_confidence"] * move["understanding_confidence"] for move in moves])

        if entry["event"] == REQUEST_EVENT:
            if is_next_entry_more_informed():
                return
            request = entry["frontend_request"]["request"]
            if "semantic_input" in request:
                interpretations = request["semantic_input"]["interpretations"]
                best_interpretation = max(interpretations, key=interpretation_confidence)
                semantic_expressions = [move["semantic_expression"] for move in best_interpretation["moves"]]
                print(InteractionTestCompiler.pretty_semantic_expressions(semantic_expressions))
            elif "natural_language_input" in request:
                ReadableNaturalLanguageMixin.handle_entry(entries, entry, index)
            elif "passivity" in request:
                print(InteractionTestCompiler.pretty_passivity())
            elif "event" in request:
                print(InteractionTestCompiler.pretty_event(request["event"]))
        if entry["event"] == RESPONSE_EVENT:
            if "output" in entry["frontend_response"]:
                moves = entry["frontend_response"]["output"]["moves"]
                print(InteractionTestCompiler.pretty_system_moves(moves))
        if entry["event"] == SELECTED_INTERPRETATION_EVENT:
            interpretation = entry["selected_interpretation"]
            if interpretation is not None:
                moves = interpretation["openqueue"]
                semantic_expressions = [move["semantic_expression"] for move in moves]
                print(InteractionTestCompiler.pretty_semantic_expressions(semantic_expressions))


class FullNaturalLanguageMixin(PrintMixin):
    @classmethod
    def handle_entry(cls, entries, entry, index):
        if entry["event"] == REQUEST_EVENT:
            request = entry["frontend_request"]["request"]
            if "natural_language_input" in request:
                hypotheses = request["natural_language_input"]["hypotheses"]
                interactiontesting_hypotheses = [(hypothesis["utterance"], hypothesis["confidence"])
                                                 for hypothesis in hypotheses]
                print(InteractionTestCompiler.pretty_hypotheses(interactiontesting_hypotheses))
            elif "semantic_input" in request:
                print(InteractionTestCompiler.pretty_semantic_input(request["semantic_input"]))
            elif "passivity" in request:
                print(InteractionTestCompiler.pretty_passivity())
            elif "event" in request:
                print(InteractionTestCompiler.pretty_event(request["event"]))
        if entry["event"] == RESPONSE_EVENT:
            if "output" in entry["frontend_response"]:
                utterance = entry["frontend_response"]["output"]["utterance"]
                print(InteractionTestCompiler.pretty_system_utterance(utterance))
        if entry["event"] == SELECTED_INTERPRETATION_EVENT:
            pass


class ReadableNaturalLanguageMixin(FullNaturalLanguageMixin):
    @classmethod
    def handle_entry(cls, entries, entry, index):
        def is_next_entry_more_informed():
            next_index = index + 1
            if len(entries) <= next_index:
                return False
            next_entry = entries[next_index]
            if next_entry["event"] == SELECTED_INTERPRETATION_EVENT:
                return next_entry["selected_hypothesis"] is not None
            return False

        if entry["event"] == REQUEST_EVENT:
            if is_next_entry_more_informed():
                return
            request = entry["frontend_request"]["request"]
            if "natural_language_input" in request:
                hypotheses = request["natural_language_input"]["hypotheses"]
                best_hypothesis = max(hypotheses, key=lambda hypothesis: hypothesis["confidence"])
                print(
                    InteractionTestCompiler.pretty_hypotheses([
                        (best_hypothesis["utterance"], best_hypothesis["confidence"])
                    ])
                )
            elif "semantic_input" in request:
                print(InteractionTestCompiler.pretty_semantic_input(request["semantic_input"]))
            elif "passivity" in request:
                print(InteractionTestCompiler.pretty_passivity())
            elif "event" in request:
                print(InteractionTestCompiler.pretty_event(request["event"]))
        if entry["event"] == RESPONSE_EVENT:
            if "output" in entry["frontend_response"]:
                utterance = entry["frontend_response"]["output"]["utterance"]
                print(InteractionTestCompiler.pretty_system_utterance(utterance))
        if entry["event"] == SELECTED_INTERPRETATION_EVENT:
            hypothesis = entry["selected_hypothesis"]
            interpretation = entry["selected_interpretation"]
            if hypothesis is not None and hypothesis["utterance"] is not None:
                hypotheses = [(hypothesis["utterance"], hypothesis["confidence"])]
                print(InteractionTestCompiler.pretty_hypotheses(hypotheses))
            elif interpretation is not None:
                ReadableSemanticMixin.handle_entry(entries, entry, index)



from tala.model.lambda_abstraction import (
    LambdaAbstractedPredicateProposition, LambdaAbstractedProposition, LambdaAbstractedGoalProposition
)
from tala.model.predicate import Predicate  # noqa
from tala.model.proposition import Proposition, PropositionSet
from tala.model.semantic_object import SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.unicodify import unicodify


class Question(SemanticObjectWithContent, AsSemanticExpressionMixin):
    TYPE_WH = "WHQ"
    TYPE_YESNO = "YNQ"
    TYPE_ALT = "ALTQ"
    TYPE_KPQ = "KPQ"
    TYPE_CONSEQUENT = "CONSEQUENT"

    TYPES = [TYPE_WH, TYPE_YESNO, TYPE_ALT, TYPE_KPQ, TYPE_CONSEQUENT]

    @classmethod
    def create_from_json_api_data(cls, question_data, included):
        class_name = question_data["type"].rsplit(".", 1)[-1]
        target_cls = globals().get(class_name)

        if target_cls is not None:
            return target_cls.create_from_json_api_data(question_data, included)
        raise Exception(f"cannot instantiate object of class {class_name}")

    def __init__(self, type_, content):
        SemanticObjectWithContent.__init__(self, content)
        self._type = type_
        self._content = content

    def __eq__(self, other):
        try:
            equality = self.type_ == other.type_ and self.content == other.content
            return equality
        except AttributeError:
            return False

    @property
    def json_api_id(self):
        if self.is_ontology_specific():
            return f"{self.ontology_name}:{self.type_}:{self.content}"
        else:
            return f"{self.type_}:{self.content}"

    @property
    def json_api_attributes(self):
        if self.is_ontology_specific() and self.ontology_name:
            return ["ontology_name", "type_"]
        else:
            return ["type_"]

    @property
    def json_api_relationships(self):
        return ["content"]

    def __hash__(self):
        return hash((self._content, self._type))

    def is_question(self):
        return True

    def is_action_question(self):
        if self.is_wh_question():
            try:
                return self.content.is_lambda_abstracted_goal_proposition()
            except AttributeError:
                return False
        elif self.is_alt_question():
            for alt in self.content:
                if (alt.is_goal_proposition() and alt.get_goal().type_ == "PERFORM_GOAL"):
                    return True
        elif self.is_yes_no_question():
            return self.content.is_goal_proposition()

    def is_wh_question(self):
        return self._type == self.TYPE_WH

    def is_yes_no_question(self):
        return self._type == self.TYPE_YESNO

    def is_alt_question(self):
        return self._type == self.TYPE_ALT

    def is_knowledge_precondition_question(self):
        return self._type == self.TYPE_KPQ

    def is_consequent_question(self):
        return self._type == self.TYPE_CONSEQUENT

    def is_understanding_question(self):
        return (self._type == self.TYPE_YESNO and self._content.is_understanding_proposition())

    def is_preconfirmation_question(self):
        return (self._type == self.TYPE_YESNO and self._content.is_preconfirmation_proposition())

    @property
    def sort(self):
        return self.predicate.sort

    @property
    def content(self):
        return self._content

    @property
    def type_(self):
        return self._type

    @property
    def predicate(self):
        return self.content.predicate

    def __str__(self):
        return "?" + unicodify(self.content)


class WhQuestion(Question):
    @classmethod
    def create_from_json_api_data(cls, lambda_abstraction_data, included):
        lambda_abstraction_entry = included.get_object_from_relationship(
            lambda_abstraction_data["relationships"]["content"]["data"]
        )
        if lambda_abstraction_entry["type"].endswith(
            LambdaAbstractedProposition.LAMBDA_ABSTRACTED_PREDICATE_PROPOSITION
        ):
            lambda_abstraction = LambdaAbstractedPredicateProposition.create_from_json_api_data(
                lambda_abstraction_entry, included
            )
            return cls(lambda_abstraction)
        if lambda_abstraction_entry["type"].endswith(LambdaAbstractedProposition.LAMBDA_ABSTRACTED_GOAL_PROPOSITION):
            return cls(LambdaAbstractedGoalProposition())
        raise NotImplementedError(f'{lambda_abstraction_entry["type"]} is not implemented')

    def __init__(self, lambda_abstraction):
        Question.__init__(self, Question.TYPE_WH, lambda_abstraction)

    def __repr__(self):
        return "WhQuestion(%r)" % self._content


class AltQuestion(Question):
    @classmethod
    def create_from_json_api_data(cls, question_data, included):
        propositions = included.get_object_from_relationship(question_data["relationships"]["content"]["data"])
        proposition_set = PropositionSet.create_from_json_api_data(propositions, included)
        return cls(proposition_set)

    def __init__(self, proposition_set):
        Question.__init__(self, Question.TYPE_ALT, proposition_set)

    def __str__(self):
        if self._contains_single_predicate():
            return "?X.%s(X), %s" % (self._predicate(), self._content)
        else:
            return Question.__str__(self)

    def _contains_single_predicate(self):
        predicates = {alt.predicate for alt in self._content if alt.is_predicate_proposition()}
        return len(predicates) == 1

    def _predicate(self):
        return list(self._content)[0].predicate


class YesNoQuestion(Question):
    @classmethod
    def create_from_json_api_data(cls, question_data, included):
        proposition_entry = included.get_object_from_relationship(question_data["relationships"]["content"]["data"])
        proposition = Proposition.create_from_json_api_data(proposition_entry, included)
        return cls(proposition)

    def __init__(self, proposition):
        Question.__init__(self, Question.TYPE_YESNO, proposition)


class KnowledgePreconditionQuestion(Question):
    def __init__(self, question):
        Question.__init__(self, Question.TYPE_KPQ, question)

    def __str__(self):
        return f"?know_answer({self.content})"


class ConsequentQuestion(Question):
    def __init__(self, lambda_abstracted_implication_proposition):
        Question.__init__(self, Question.TYPE_CONSEQUENT, lambda_abstracted_implication_proposition)

    def get_embedded_consequent_question(self):
        consequent_predicate = self.content.consequent_predicate
        lambda_abstracted_consequent_proposition = LambdaAbstractedPredicateProposition(
            consequent_predicate, consequent_predicate.ontology_name
        )
        return WhQuestion(lambda_abstracted_consequent_proposition)



from tala.ddd.loading.extended_ddd_loader import ExtendedDDDLoader
from tala.ddd.loading.ddd_set_loader import DDDSetLoader
from tala.utils.chdir import chdir


class ExtendedDDDSetLoader(DDDSetLoader):
    def __init__(self, ddd_manager, overridden_ddd_config_paths=None):
        super(ExtendedDDDSetLoader, self).__init__(overridden_ddd_config_paths)
        self._ddd_manager = ddd_manager

    def ddds_as_list(self, ddds, path=".", *args, **kwargs):
        with chdir(path):
            ddds_dict = self._load_ddds(ddds, *args, **kwargs)
            return list(ddds_dict.values())

    def _load_ddds(self, ddd_names, *args, **kwargs):
        ddds = {}
        configs = {}
        for ddd_name in ddd_names:
            configs[ddd_name] = self._ddd_config(ddd_name)

        for ddd_name in ddd_names:
            config = configs[ddd_name]
            ddds[ddd_name] = self._load_ddd(ddd_name, config, *args, **kwargs)
        return ddds

    def _load_ddd(self, *args, **kwargs):
        ddd_loader = ExtendedDDDLoader(self._ddd_manager, *args, **kwargs)
        return ddd_loader.load()

    def ensure_ddds_loaded(self, ddds, path=".", *args, **kwargs):
        with chdir(path):
            ddds = self._load_ddds(ddds, *args, **kwargs)
            for ddd in list(ddds.values()):
                self._ddd_manager.ensure_ddd_added(ddd)
